/*
 * hytelnet.c - a cgi gateway to the hytelnet system
 *
 *	This program is based on the unix hytelnet software developed
 *	by Earl Fogel at University of Saskatchewan.
 *
 *	Written for cgi by Ken Guyre - 7/6/94
 *      Additional mods by Earl Fogel - ??/09/94
 */
#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>	/* some systems require <strings.h> instead */

#define UNIX
#define TRUE	1
#define FALSE	0

/*********************************
 * Things you may want to adjust *
 *********************************/

#define STARTDIR "/usr/local/hytelnet"
#define TELNET_COMMAND "telnet"
#define TN3270_COMMAND "tn3270"
#define RLOGIN_COMMAND "rlogin"
#define FTP_COMMAND    "ftp"
#define PAD_COMMAND    "pad"

char *unzip_command="unzip -aj";

/*
 * The string that separates an Internet address and port number.
 *
 *    " "       - most systems            (eg. telnet 130.132.1.4 300)
 *    " /port=" - VMS Multinet systems    (eg. telnet 130.132.1.4 /port=300)
 *    "#"       - DOS NCSA Telnet systems (eg. telnet 130.132.1.4#300)
 */
#define PORTSEP " "		/* most systems */

int connect_by_name = TRUE;	/* TRUE to use Internet name for telnet */
int connect_by_number = TRUE;	/* TRUE to use Internet number for telnet */

extern char *getenv(), *strcpy(), *strncpy();	/* comment this out on IBM RS 6000 */

/*********************************************************
 * You shouldn't need to change anything below this line *
 *********************************************************/

#include <sys/types.h>
#include <sys/stat.h>

#define TELNET    1
#define TN3270    2
#define RLOGIN    3
#define FTP       5
#define PAD       6
#define GENERIC   7
#define NCOMMANDS 7
int command_ok[NCOMMANDS+1] = {0};
char command[16];

#define MAXFNAME 80	/* max length of filename and other stuff in <>'s */
#define STARTFILE "start"

#define STREQ(a,b) (strcasecmp(a,b) == 0)
#define STRNEQ(a,b,c) (strncasecmp(a,b,c) == 0)
#define printable(c) (((c)>31 && (c)<=127) || (c)==9 || (c)==10)

#define ON	1
#define OFF	0
#define HLINE   196     /* horizontal line segment */
#define VLINE   179     /* vertical line segment */
#define VLINE2  186     /* vertical line segment */
#define UL      218     /* upper left corner */
#define UR      191     /* upper right corner */
#define LL      192     /* lower left corner */
#define LR      217     /* lower right corner */
#define UL2     214     /* upper left corner */
#define UR2     183     /* upper right corner */
#define LL2     211     /* lower left corner */
#define LR2     189     /* lower right corner */
#define BULLET  4       /* list item marker */
#define BLOT1   178
#define BLOT2   177
#define BLOT3   176

char *startfile = STARTFILE;
char *startdir  = STARTDIR;

char *interaddr(), *findport(), *nextword();
char *mystrncpy(), *strcasestr(), *strsave(), *match_fgets();
FILE *myfopen(), *htmlfopen();

typedef struct {
	char name[128];
	char val[128];
} entry;

main()
{
    entry file;
    char *cl, filename[MAXFNAME];

    command_ok[TELNET]  = (TELNET_COMMAND != NULL);
    command_ok[TN3270]  = (TN3270_COMMAND != NULL);
    command_ok[RLOGIN]  = (RLOGIN_COMMAND != NULL);
    command_ok[FTP]     = (FTP_COMMAND    != NULL);
    command_ok[PAD]     = (PAD_COMMAND    != NULL);
    command_ok[GENERIC] = FALSE;

    printf("Content-type: text/html\n\n");
    cl = getenv("QUERY_STRING");

    /* the first invocation gets the startfile */
    if (!cl || !cl[0])
	strcpy(filename, startfile);
    else {
	getword(file.val, cl, '&');
	plustospace(file.val);
	unescape_url(file.val);
	getword(file.name, file.val, '=');
	strcpy(filename, file.val);
    }
    present_file(filename);
}


/*
 * Read the file and convert to HTML
 */

present_file(name)
char *name;
{
    FILE *fp;
    int firstline = 1;
    char in[182], out[182];
    char address[80], telanchor[80], title[80];
    char *anchor = "<A HREF=\"/cgi-bin/hytelnet?file=";
    char *cp, *op, *cp2, *cp3, *acp;
    unsigned char uc;

    /*
     * if there's already an html version, send it as is
     */
    if ((fp = htmlfopen(name, "r"))) {
	while (fgets(in, 181, fp)) {
	    fputs(in,stdout);
	}
	return;
    } else if ((fp = myfopen(name, "r")) == NULL) {
	printf("%s is a bad hypertext link\n", name);
	exit(1);
    }

    while (fgets(in, 181, fp)) {
	for (cp=in+strlen(in)-1; *cp == 13 || *cp == 10; *cp--='\0');

	if (firstline){
	    for (cp=in; isspace(*cp) && *cp!='\0'; cp++);
	    if (*cp == '\0')
		continue;
	    strcpy(title, cp);
	    printf("<title>Hytelnet - %s</title>\n<PRE>\n", title);
	    firstline = 0;
	}
	
	for (cp=in, op=out; *cp != '\0'; cp++) {
	    uc = *cp;
	    if (!printable(uc)) {
		if      (uc == HLINE)	*op = '-';
		else if (uc == VLINE)	*op = '|';
		else if (uc == VLINE2)	*op = '|';
		else if (uc == UL)	*op = '+';
		else if (uc == UR)	*op = '+';
		else if (uc == LL)	*op = '+';
		else if (uc == LR)	*op = '+';
		else if (uc == UL2)	*op = '+';
		else if (uc == UR2)	*op = '+';
		else if (uc == LL2)	*op = '+';
		else if (uc == LR2)	*op = '+';
		else if (uc == BULLET)	*op = '*';
		else if (uc == BLOT1)	*op = '*';
		else if (uc == BLOT2)	*op = '*';
		else if (uc == BLOT3)	*op = '*';
		else {
		    *op = ' ';
		}
		op++;
	    } else if (*cp == '\\') {
		cp2 = cp+1;
		if (*cp2 == '<' || iscommand(cp2))
		    cp++;
		if (*cp == '<') {
		    *op++ = '&';
		    *op++ = 'l';
		    *op++ = 't';
		    *op++ = ';';
		} else
		    *op++ = *cp;
	    } else if (*cp == '<') {	/* start of link? */
		for (cp2 = cp+1;
		    *cp2 != '>' && *cp2 != '<' && *cp2 != '\0' &&
		    cp2-cp+1 < MAXFNAME; cp2++)
		    ; /* NULL BODY */
		if (*cp2 == '>') { /* it's a link */
		    for (cp3 = anchor; *cp3; *op++ = *cp3++);
		    for (cp3=cp+1; cp3<cp2; *op++ = *cp3++);
		    *op++ = '"';
		    *op++ = '>';
		    *op++ = '&';
		    *op++ = 'l';
		    *op++ = 't';
		    *op++ = ';';
		    for(++cp; cp < cp2; *op++ = *cp++);
		    *op++ = '&';
		    *op++ = 'g';
		    *op++ = 't';
		    *op++ = ';';
		    *op++ = '<';
		    *op++ = '/';
		    *op++ = 'a';
		    *op++ = '>';
		} else {
		    *op++ = *cp;
		}
	    } else if (cp == in || isspace(*(cp-1))) {  /* a command? */
		if (iscommand(cp)) {
		    acp=interaddr(nextword(cp),address);
		    while (acp != NULL) {
			for (;cp<acp; *op++ = *cp++);
			sprintf(telanchor, "<A HREF=\"%s://%s\">",
				command, address);
			for (cp2=telanchor; *cp2!='\0'; *op++=*cp2++);
			acp += strlen(address);
			for (;cp<acp; *op++=*cp++);
		        *op++ = '<';
		        *op++ = '/';
		        *op++ = 'a';
		        *op++ = '>';
			acp=interaddr(acp,address);
		    }
		}
		*op++ = *cp;
	    } else if (*cp == '>') {
		    *op++ = '&';
		    *op++ = 'g';
		    *op++ = 't';
		    *op++ = ';';
	    } else if (*cp == '&') {
		    *op++ = '&';
		    *op++ = 'a';
		    *op++ = 'm';
		    *op++ = 'p';
		    *op++ = ';';
	    } else {
		*op++ = *cp;
	    }
	}
	*op = '\0';
	printf("%s\n", out);
    }
    printf("</pre>\n");
}


/*
 * Check if string cp begins with an acceptable command, and, if so,
 * return the command type.
 */
iscommand(cp)
char *cp;
{
    if (command_ok[TELNET] && STRNEQ(cp,"TELNET ",7)) {
	strcpy(command, "telnet");
	return(TELNET);
    }
    if (command_ok[TN3270] && STRNEQ(cp,"TN3270 ",7)) {
	strcpy(command, "tn3270");
	return(TN3270);
    }
    if (command_ok[RLOGIN] && STRNEQ(cp,"RLOGIN ",7)) {
	strcpy(command, "rlogin");
	return(RLOGIN);
    }
    if (command_ok[FTP] && STRNEQ(cp,"FTP ",4)) {
	strcpy(command, "ftp");
	return(FTP);
    }

    return(0);
}

/*
 * Search through a string for the first valid internet address.
 * We define an internet address as a word consisting of alpha-
 * numeric characters and dots (and with at least one embedded dot).
 *
 * If the address is immediately followed by a space and a port number,
 * or a " -l " and a login id, those are included as well.
 *
 * This is not a general purpose routine.
 *
 * Returns an pointer to the address if one is found, NULL otherwise.
 */
char *
interaddr(cp, buf)
char *cp, *buf;
{
    int isaddr, isname;
    char *cp2, *word;

    isaddr = FALSE;
    isname = FALSE;

    while (isspace(*cp)) cp++;		/* skip spaces */
    if (STRNEQ(cp, "or ", 3)) {		/* skip the word 'or' */
	cp += 3;
	while (isspace(*cp)) cp++;
    }

    word = cp;				/* start of word? */
    while (*cp && (strchr("\"-_=.@", *cp) || isalnum(*cp))) {
	if (*cp == '.' && isalnum(*(cp+1)) )
	    isaddr = TRUE;
	if (isalpha(*cp))
	    isname = TRUE;
	cp++;
    }

    if (isaddr) {			/* it's an internet address */
	if ((connect_by_name && isname) ||
	    (connect_by_number && !isname) ) {
	    if (STRNEQ(cp," -l ",4)) {	/* -l argument to rlogin */
		cp = nextword(cp);
		cp = nextword(cp);
		while (isalnum(*cp)) cp++;
	    } else if (isspace(*cp)) {	/* check for port number */
		cp2 = findport(cp);
		if (cp2) {
		    cp = cp2;
		    while (isdigit(*cp)) cp++;
		}
	    }
	    while (!isalnum(*cp) && *cp != '"') cp--;
	    mystrncpy(buf,word,cp-word+1);
	    /* printw("interaddr: '%s'\n", buf); refresh(); */
	    return(word);
	}
    }
    return(NULL);
}

/*
 * Looks for something resembling a port number in a telnet argument string
 */
char *
findport(cp)
char *cp;
{
    char *port;

    cp = port = nextword(cp);
    while (isdigit(*cp)) cp++;
    if (cp>port && (isspace(*cp) || *cp=='\0')) {
	return(port);
    }

    return(NULL);
}

/*
 * Return pointer to the next word in a string
 */
char *
nextword(cp)
char *cp;
{
    while (*cp != ' ' && *cp != '\0') cp++;	/* skip non-spaces */
    while (isspace(*cp)) cp++;			/* skip spaces */
    return(cp);
}

/*
 * Copy the next word in s2 to s1
 */
copyword(s1,s2)
char *s1, *s2;
{
    while (*s2 != ' ' && *s2 != '\0')
	*s1++ = *s2++;
    *s1 = '\0';
}


/*
 * my strncpy() terminates strings with a null byte.
 * Writes a null byte into the n+1 byte of dst.
 */
char *
mystrncpy(dst, src, n)
char *dst, *src;
int n;
{
    char *val;

    val = strncpy(dst, src, n);
    *(dst+n) = '\0';
    return val;
}


/*
 * open a file relative to startdir
 * convert filename to lowercase and, if not found,
 * look in subdir (eg. look for US001 as us0/us001)
 */
FILE *
myfopen(fname,type)
char *fname, *type;
{
    char *cp, buf[200];
    FILE *fp;

    for (cp=fname; *cp!='\0'; cp++) {
	if (isupper(*cp))
	    *cp = tolower(*cp);
    }

    sprintf(buf, "%s/%s", startdir, fname);
    fp = fopen(buf,type);
    if (fp == NULL) {			/* try without .html */
	strcpy(buf, startdir);
	strcat(buf, "/");
	strncat(buf, fname, 3);
	strcat(buf, "/");
	strcat(buf, fname);
	fp=fopen(buf, type);
    }
    return(fp);
}

/*
 * as above, but append .html to the filename
 */
FILE *
htmlfopen(fname,type)
char *fname, *type;
{
    char *cp, buf[200];
    FILE *fp;

    for (cp=fname; *cp!='\0'; cp++) {
	if (isupper(*cp))
	    *cp = tolower(*cp);
    }

    sprintf(buf, "%s/%s.html", startdir, fname); /* try .html version */
    fp = fopen(buf,type);
    if (fp == NULL) {			/* look in subdirectory */
	strcpy(buf, startdir);
	strcat(buf, "/");
	strncat(buf, fname, 3);
	strcat(buf, "/");
	strcat(buf, fname);
	strcat(buf, ".html");
	fp=fopen(buf, type);
    }
    return(fp);
}

char *
strcasestr(s1,s2)
char *s1, *s2;
{
    int n;

    n = strlen(s2);

    while (*s1 && strncasecmp(s1,s2,n) != 0) {
	s1++;
    }

    return (*s1)? s1: NULL;
}

char *
strsave(s)
char *s;
{
    char *cp;
    char *malloc();

    cp = (char *) malloc((unsigned)strlen(s)+1);
    if (cp) strcpy(cp, s);
    return(cp);
}

